package org.lossywav.algorithm;

//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007, 2008 Nick Currie, Copyleft.
//    Copyright (C) 2008, C# port by Greg Chudov.
//    Copyright (C) 2008, Java port by Josep Maria Antoln
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> yahoo <dot> com
import java.util.Arrays;

import org.lossywav.algorithm.SpreadingRecord.SpreadingNew;
import org.lossywav.dsp.ComplexBuffer;
import org.lossywav.dsp.InplaceFFT;
import org.lossywav.dsp.WindowFunction;
import org.lossywav.dsp.WindowingType;
import org.lossywav.util.CircularBuffer4;
import org.lossywav.util.Constants;
import org.lossywav.util.ReadBlockPosition;

public class Analyzer {
	private transient final InplaceFFT fft;

	// Calculated tables dependend on input.
	private transient float[][] scaledWindowTable;
	private transient final SpreadingRecord[] spreadingRecordOld;
	private transient final SpreadingNew[] spreadingNew;
	private transient final FFTRecord[] fftRecords;
	private transient double[] skewing_function;
	/**
	 * Temporary array used in spread_complex function. It is stored as a member
	 * to avoid to recreate it each time the function is called
	 */
	private transient double[] spreadComplexResult;

	// Input/Output
	private int[][] resultBlock;
	private int[][] correctionBlock;

	// Parameters
	private double snrValue;
	private double signalToNoiseRatioInAmplitude;
	private double noiseThresholdShift;
	private double noiseThresholdShiftInAmplitude;
	private double qualityBasedMinimumBitsToKeep;
	private int maximumClipsPerChannel;
	private final boolean[] activatedAnalysis = new boolean[Constants.PRECALC_ANALYSES];
	private int analysisMode;
	private double upperFftLimit;
	private int sampleRate;
	private int codecBlockSize;
	private double scalingFactor;
	private double shapingFactor;
	private boolean sortSpread;

	// Internal helper variables
	private transient boolean shapingIsOn;
	private transient double[] shapingA;
	private transient double[] shapingB;
	private transient double[] fifo;
	private transient double[][] saved_fifo_state;
	private transient int positiveClip;
	private transient int negativeClip;
	private transient short maxAllowedBitsToRemove;
	private transient int maxAnalysisBitLength;

	private transient FFTStatus[][] savedFFTResults;
	private transient final FFTStatus minimumFFTResult = new FFTStatus();
	private transient final FFTStatus currentFFTResult = new FFTStatus();

	/**
	 * Constructor.
	 */
	public Analyzer() {
		fft = new InplaceFFT();

		spreadingNew = new SpreadingNew[Constants.SPREADING_STEPS + 1];
		for (int i = 0; i <= Constants.SPREADING_STEPS; i++) {
			final float width = (float)i / Constants.SPREADING_STEPS;
			spreadingNew[i] = new SpreadingNew();
			spreadingNew[i].width = width;
			spreadingNew[i].divisor = 1.0f / (1.0f + 2.0f * width);
		}

		scalingFactor = 1.0;

		spreadingRecordOld = new SpreadingRecord[Constants.PRECALC_ANALYSES];
		// Max +1 in order to access it from 1 to MAX instead of 0 to MAX-1
		fftRecords = new FFTRecord[Constants.MAX_FFT_BITWIDTH + 1];
	}

	public int[][] getResultBlock() {
		return resultBlock;
	}

	public int[][] getCorrectionBlock() {
		return correctionBlock;
	}

	public double getScalingFactor() {
		return scalingFactor;
	}

	public void setScalingFactor(double scalingFactor) {
		this.scalingFactor = scalingFactor;
	}

	public double getShapingFactor() {
		return shapingFactor;
	}

	public void setShapingFactor(double shapingFactor) {
		this.shapingFactor = shapingFactor;
		shapingIsOn = shapingFactor != 0;
	}

	public double getSnrValue() {
		return snrValue;
	}

	public void setSnrValue(double snrValue) {
		this.snrValue = snrValue;
		signalToNoiseRatioInAmplitude = Math.pow(10.0, snrValue / 20.0);
	}

	public double getNoiseThresholdShift() {
		return noiseThresholdShift;
	}

	public void setNoiseThresholdShift(double noiseThresholdShift) {
		this.noiseThresholdShift = noiseThresholdShift;
		noiseThresholdShiftInAmplitude = Math.pow(10.0, noiseThresholdShift / 20.0);
	}

	public double getQualityBasedMinimumBitsToKeep() {
		return qualityBasedMinimumBitsToKeep;
	}

	public void setQualityBasedMinimumBitsToKeep(double dynamicMinimumBitsToKeep) {
		qualityBasedMinimumBitsToKeep = dynamicMinimumBitsToKeep;
	}

	public int getMaximumClipsPerChannel() {
		return maximumClipsPerChannel;
	}

	public void setMaximumClipsPerChannel(int maximumClipsPerChannel) {
		this.maximumClipsPerChannel = maximumClipsPerChannel;
	}

	public int getAnalysisMode() {
		return analysisMode + Constants.MIN_PARAM_ANALYSIS;
	}

	public void setAnalysisMode(int analysisMode) {
		this.analysisMode = analysisMode - Constants.MIN_PARAM_ANALYSIS;
	}

	public boolean isFFT32() {
		return activatedAnalysis[0];
	}

	public void setFFT32(boolean impulse) {
		activatedAnalysis[0] = impulse;
	}

	/**
	 * getter of upperFftLimit
	 * 
	 * @return the upperFftLimit
	 */
	public double getUpperFftLimit() {
		return upperFftLimit;
	}

	/**
	 * setter of upperFftLimit
	 * 
	 * @param upperFftLimit
	 *            the upperFftLimit to set
	 */
	public void setUpperFftLimit(double upperFftLimit) {
		this.upperFftLimit = upperFftLimit;
	}

	/**
	 * getter of sortSpread
	 * 
	 * @return the sortSpread
	 */
	public boolean isSortSpread() {
		return sortSpread;
	}

	/**
	 * setter of sortSpread
	 * 
	 * @param sortSpread
	 *            the sortSpread to set
	 */
	public void setSortSpread(boolean sortSpread) {
		this.sortSpread = sortSpread;
	}

	/**
	 * Prepares the different structures before calling process for the first
	 * time. Call it after having set the different public parameters.
	 * 
	 * @param channels
	 *            number of channels of the input source.
	 * @param bitsPerSample
	 *            bits per sample of the input source.
	 * @param sampleRate
	 *            sampling rate of the input source.
	 */
	public void initialize(int channels, int bitsPerSample, int sampleRate, int codecBlockSize, double frequencyLimit,
			int underlap) {
		this.codecBlockSize = codecBlockSize;
		this.sampleRate = sampleRate;

		// Max +1, since we want to go from 1 to max, not from 0 to max-1
		scaledWindowTable = new float[codecBlockSize * 2 + 1][];
		spreadComplexResult = new double[codecBlockSize + 2];

		resultBlock = new int[channels][codecBlockSize];
		correctionBlock = new int[channels][codecBlockSize];

		fifo = new double[Constants.SHAPING_LENGTH + codecBlockSize];
		saved_fifo_state = new double[channels][Constants.SHAPING_LENGTH + codecBlockSize];

		savedFFTResults = new FFTStatus[channels][Constants.PRECALC_ANALYSES];

		positiveClip = (1 << (bitsPerSample - 1)) - 1;
		negativeClip = -1 - positiveClip;

		maxAllowedBitsToRemove = (short) (bitsPerSample - Constants.MINIMUM_BITS_TO_KEEP);

		if (sampleRate > 46050) {
			shapingA = Constants.order_4_48000_a.clone();
			shapingB = Constants.order_4_48000_b.clone();
		} else {
			shapingA = Constants.order_4_44100_a.clone();
			shapingB = Constants.order_4_44100_b.clone();
		}
		double sf = 1.0;
		for (int i = 0; i < Constants.SHAPING_LENGTH; i++) {
			sf *= shapingFactor;
			shapingA[i] *= sf;
			shapingB[i] *= sf;
		}
		for (int i = 0; i < channels; i++) {
			for (int j = 0; j < Constants.PRECALC_ANALYSES; j++) {
				savedFFTResults[i][j] = new FFTStatus();
				savedFFTResults[i][j].start = -1;
			}
		}		
		
		// Enable parameter-specific analysis
		for (int i = 0; i <= analysisMode; i++) {
			// activatedAnalysis[0] is Impulse
			activatedAnalysis[i + 1] = true;
		}
		// Enable 1024 FFT analysis. FFT32 one is enabled via the setFFT32
		// method.
		activatedAnalysis[Constants.PRECALC_ANALYSES - 1] = true;

		fillScaledWindow((float) scalingFactor);
		precalculateAnalysis(frequencyLimit, underlap);
	}

	public void process(int channel, CircularBuffer4 rotatingBlocks, ChannelStatus channelStatus, LossyWavStatus status) {

		channelStatus.maximumBitsToRemove = (short) Math.min(maxAllowedBitsToRemove, Math.max(0, 
				(int) Math.floor(channelStatus.currentBits - qualityBasedMinimumBitsToKeep)));

		resetCurrentFFTResult(channelStatus);
		resetMinimumFFTResult(channelStatus);

		// TODO: This if is not present in delphi
		if (channelStatus.maximumBitsToRemove == 0) {
			for (int analysis_number = 0; analysis_number < Constants.PRECALC_ANALYSES; analysis_number++) {
				savedFFTResults[channel][analysis_number].copyFrom(minimumFFTResult);
			}
		} else {
			final int numSamplesTwoPrev = rotatingBlocks.getLength(ReadBlockPosition.TWOPREVIOUS.index);
			final int numSamplesPrev = rotatingBlocks.getLength(ReadBlockPosition.PREVIOUS.index);
			final int numSamplesCurrent = rotatingBlocks.getLength(ReadBlockPosition.CURRENT.index);
			final int numSamplesNext = rotatingBlocks.getLength(ReadBlockPosition.NEXT.index);
			final int neg_codec_block_start = -(numSamplesTwoPrev + numSamplesPrev);

			for (short currentAnalysis = 0; currentAnalysis < Constants.PRECALC_ANALYSES; currentAnalysis++) {
				final SpreadingRecord anRecord = spreadingRecordOld[currentAnalysis];
				final int analysisNumBits = anRecord.fft_bit_length;
				final int analysisNumSamples = 1 << analysisNumBits;

				// TODO: Shouldn't always analysisNumSamples be <=
				// codec_block_size*2 ?
				if (activatedAnalysis[currentAnalysis] && analysisNumSamples <= codecBlockSize * 2) {

					final FFTRecord anFftRecord = fftRecords[analysisNumBits];
					final int pos_codec_block_start = numSamplesCurrent + numSamplesNext - analysisNumSamples;
					final int shiftbits = maxAnalysisBitLength - analysisNumBits;

					for (int currentAnaBlock = 0; currentAnaBlock <= anFftRecord.analysis_blocks; currentAnaBlock++) {

						int sampleBlockStart = (int) Math.floor(anFftRecord.actual_analysis_blocks_start
								+ currentAnaBlock * anFftRecord.fft_underlap_length);
						sampleBlockStart = Math.min(Math.max(sampleBlockStart, neg_codec_block_start),
								pos_codec_block_start);

						if (currentAnaBlock == 0) {
							if (sampleBlockStart == savedFFTResults[channel][currentAnalysis].start) {
								currentFFTResult.copyFrom(savedFFTResults[channel][currentAnalysis]);
								if (currentFFTResult.btr < minimumFFTResult.btr || minimumFFTResult.btr == -1) {
									minimumFFTResult.copyFrom(currentFFTResult);
								}
							}
						} else {
							if (fillFFTInput(rotatingBlocks, sampleBlockStart, analysisNumSamples, analysisNumBits,
									channel) > 0.0) {

								fft.fftReal(analysisNumBits - 1);

								double spreaddBs;
								if (sortSpread) {
									spreaddBs = calculateMinimumdBSort(anFftRecord, shiftbits);

								} else {
									spreaddBs = calculateMinimumdBNew(anFftRecord, spreadingRecordOld[currentAnalysis],
											shiftbits);
								}
								spreaddBs = Math.min(Constants.THRESHOLD_INDEX_SPREAD_MUL - 1, Math.max(0, spreaddBs))
										* Constants.THRESHOLD_INDEX_SPREAD;
								currentFFTResult.btr = anRecord.thresholdIndex[(int) Math.floor(spreaddBs)];
							} else {
								currentFFTResult.btr = channelStatus.maximumBitsToRemove;
							}
							if (currentFFTResult.btr < minimumFFTResult.btr || minimumFFTResult.btr == -1) {
								minimumFFTResult.copyFrom(currentFFTResult);
							}
							if (currentAnaBlock == anFftRecord.analysis_blocks) {
								currentFFTResult.start = (short) (sampleBlockStart - codecBlockSize);
								savedFFTResults[channel][currentAnalysis].copyFrom(currentFFTResult);
							}
						}
					}
				}
			}
			minimumFFTResult.btr = (short) Math.max(minimumFFTResult.btr, 0);
			 // prepare FIFO buffer with saved state information for this channel;
			 for ( int j=0; j < Constants.SHAPING_LENGTH ; j++ ) {
			    fifo[j] = saved_fifo_state[channel][j];
			 }
			 
			remove_bits(channel, minimumFFTResult.btr, channelStatus, rotatingBlocks);
			/* This is what the Delphi code does, to accomodate for the linkChannels outside of Analysis.process
			 * // save FIFO state information to temp storage 
			 * for ( int j = 0; j < Constants.SHAPING_LENGTH; j++ ) { 
			 * temp_fifo_state[channel][j] = fifo[j + codecBlockSize];
			 * }
			 * }
			 * 
			 */
             // copy FIFO saved state from temp storage to saved storage, i.e. no more bit-removal for this codec-block.
		     for ( int j = 0; j < Constants.SHAPING_LENGTH; j++ ) {
		        saved_fifo_state[channel][j] = fifo[j + codecBlockSize];
		     }
		}
	}

	/**
	 * Fills each table with the generated window.
	 * 
	 * @param scaling
	 *            amount to scale the window (1.0f = no scaling)
	 */
	private void fillScaledWindow(float scaling) {
		final WindowFunction window = new WindowFunction();
		window.setWindowType(WindowingType.HANN);
		for (int bits = 1; bits <= Constants.MAX_FFT_BITWIDTH; bits++) {
			scaledWindowTable[bits] = window.generate(1 << bits, scaling);
		}
	}

	private void resetCurrentFFTResult(ChannelStatus channelStatus) {
		currentFFTResult.start = -9999;
		currentFFTResult.btr = channelStatus.maximumBitsToRemove;
	}

	private void resetMinimumFFTResult(ChannelStatus channelStatus) {
		minimumFFTResult.start = -9999;
		minimumFFTResult.btr = channelStatus.maximumBitsToRemove;
	}

	public void remove_bits(int channel, short bits_to_remove_from_this_channel, ChannelStatus channelStatus,
			CircularBuffer4 audioBlock) {

		int this_two_power = 1 << bits_to_remove_from_this_channel;
		double this_two_power_recip  = 1.0 /(double) this_two_power;
		short this_channel_clips = 999;
		// This while loop enforces the process to be started again, if the
		// allowed number
		// of clips have been exceeded and it is still possible to reduce the
		// bits to remove.
		// TODO: Warning! there's a problem with this logic. if it can't avoid
		// clipping, no output is generated
		while (bits_to_remove_from_this_channel >= 0 && this_channel_clips > maximumClipsPerChannel) {
			this_channel_clips = 0;

			final int max_sample = positiveClip >> bits_to_remove_from_this_channel << bits_to_remove_from_this_channel;

			if (shapingIsOn) {
				double fifo_multiplier_multiplier;
				if (bits_to_remove_from_this_channel == 0) {
					fifo_multiplier_multiplier = 0.5;
				} else {
					fifo_multiplier_multiplier = 1.0;
				}
				double fifo_multiplier = 1;
				for (int i = 0; i < audioBlock.getCurrentLength(); i++) {
					final int sample = (int) Math.floor(audioBlock.getChannelSample(channel, i) * scalingFactor);
					// Position i + 6 is the position for the current sample.
					// Position i is 6 positions before.
					final double wanted_temp = sample + 
						fifo[i + 5] * shapingB[0] + fifo[i + 4] * shapingB[1] +
						fifo[i + 3] * shapingB[2] + fifo[i + 2] * shapingB[3] +
						fifo[i + 1] * shapingB[4] + fifo[i] * shapingB[5];

					final double output_temp = Math.round(wanted_temp*this_two_power_recip)*this_two_power;

					fifo[i + 6] = (output_temp - wanted_temp - 
							fifo[i + 5] * shapingA[0] - fifo[i + 4] * shapingA[1] -
							fifo[i + 3] * shapingA[2] - fifo[i + 2] * shapingA[3] -
							fifo[i + 1] * shapingA[4] - fifo[i]		* shapingA[5])
							* fifo_multiplier;

					fifo_multiplier = fifo_multiplier * fifo_multiplier_multiplier;

					int new_sample = (int)Math.round(output_temp);
					if (new_sample > max_sample) {
						new_sample = max_sample;
						this_channel_clips++;
						if (this_channel_clips > maximumClipsPerChannel) {
							channelStatus.bitsLost++;
							bits_to_remove_from_this_channel--;
							break;
						}
					} else if (new_sample < negativeClip) {
						new_sample = negativeClip;
						this_channel_clips++;
						if (this_channel_clips > maximumClipsPerChannel) {
							channelStatus.bitsLost++;
							bits_to_remove_from_this_channel--;
							break;
						}
					}
					resultBlock[channel][i] = new_sample;
					correctionBlock[channel][i] = sample - (int) Math.round(new_sample / scalingFactor);
				}
			} else {
				for (int i = 0; i < audioBlock.getCurrentLength(); i++) {
					// The scaling is done here, because scaling is applied to
					// the scanned data via the window function.
					// This compensates for it.
					final int sample = (int) (audioBlock.getChannelSample(channel, i) * scalingFactor);
					int new_sample = (int) Math.round(sample*this_two_power_recip)*this_two_power;
					if (new_sample > max_sample) {
						new_sample = max_sample;
						this_channel_clips++;
						if (this_channel_clips > maximumClipsPerChannel) {
							channelStatus.bitsLost++;
							bits_to_remove_from_this_channel--;
							break;
						}
					} else if (new_sample < negativeClip) {
						new_sample = negativeClip;
						this_channel_clips++;
						if (this_channel_clips > maximumClipsPerChannel) {
							channelStatus.bitsLost++;
							bits_to_remove_from_this_channel--;
							break;
						}
					}
					resultBlock[channel][i] = new_sample;
					correctionBlock[channel][i] = sample - (int) Math.round(new_sample / scalingFactor);
				}
			}
			channelStatus.clippedSamples = this_channel_clips;
			channelStatus.bitsToRemove = bits_to_remove_from_this_channel;
		}
		if (bits_to_remove_from_this_channel < 0) {
			channelStatus.bitsToRemove = 0;
		}
	}

	/**
	 * Precalculates the values of analysis for the predefined bit lengths and
	 * sampleRate. Cannot be made static, needs several parameters.
	 */
	private void precalculateAnalysis(double frequencyLimit, int underlap) {
		final int codecBlockBits = (int) Math.round(Math.log10(codecBlockSize) / Math.log10(2));
		for (int bitwidth = 0; bitwidth <= Constants.MAX_FFT_BITWIDTH; bitwidth++) {
			fftRecords[bitwidth] = new FFTRecord();
			fftRecords[bitwidth].generate(bitwidth, sampleRate, frequencyLimit, codecBlockSize, underlap);
		}
		for (int analysisIndex = 0; analysisIndex < Constants.PRECALC_ANALYSES; analysisIndex++) {
			final int currentBitLength = codecBlockBits + Constants.PRECALC_ANALYSIS_BITS_DIFF[analysisIndex];
			spreadingRecordOld[analysisIndex] = new SpreadingRecord();
			spreadingRecordOld[analysisIndex].generate(analysisIndex, currentBitLength, frequencyLimit,
					fftRecords[currentBitLength]);
		}

		maxAnalysisBitLength = codecBlockBits + Constants.PRECALC_ANALYSIS_BITS_DIFF[Constants.PRECALC_ANALYSES - 1];
		skewing_function = new double[codecBlockSize + 2];
		skewing_function[0] = 0;
		final double min_bin_width = fftRecords[maxAnalysisBitLength].bin_width;
		final double sa_lfb = Math.log10(Constants.LOW_FREQ_LIMIT); // 20Hz lower limit for skewing;
		final double sa_mfb = Math.log10(Constants.MID_FREQ_BAND_SKEWING); // 3445.3125Hz upper limit for skewing;
		final double sa_dfb = sa_mfb - sa_lfb; // skewing range;

		for (int i = 1; i <= fftRecords[maxAnalysisBitLength].hi_bin + 1; i++) {
			final double sa_tfb = Math.log10(i * min_bin_width);
			if (sa_tfb <= sa_mfb) {
				skewing_function[i] = (double) Math.pow(10, (Math.pow(Math.sin(Constants.HALF_PI
						* Math.max(0, sa_tfb - sa_lfb) / sa_dfb), 0.75) - 1)
						* Constants.SKEWING_AMPLITUDE * 0.05);
			} else {
				skewing_function[i] = 1;
			}
		}
	}

	private double fillFFTInput(CircularBuffer4 rotatingBlocks, int currentBlockStart, int fftLength, int fftbits,
			int channel) {
		double rmsValue = 0;
		final ComplexBuffer xRef = fft.getX();

		final int startIdx = ReadBlockPosition.CURRENT.index * codecBlockSize + currentBlockStart;

		final float[] fftWindowLookup = scaledWindowTable[fftbits];
		final int loopEnd = fftLength >> 1;
		for (int index = 0; index < loopEnd; index++) {
			final int sampleIdx = 2 * index + startIdx;
			final double sample1 = rotatingBlocks.getCircularSample(channel, sampleIdx);
			final double sample2 = rotatingBlocks.getCircularSample(channel, sampleIdx + 1);

			// No explicit sample scaling because the window is already scaled.
			xRef.setComplex(index, sample1 * fftWindowLookup[2 * index], sample2 * fftWindowLookup[2 * index + 1]);
			rmsValue += sample1 * sample1 + sample2 * sample2;
		}
		return Math.sqrt(rmsValue / fftLength);
	}

	public double calculateMinimumdBSort(FFTRecord fftRecord, int shiftbits) {
		final double spread_average = postProcessFFTResults(fftRecord, shiftbits);

		final int j = fftRecord.lo_bin;
		Arrays.sort(spreadComplexResult, j, fftRecord.hi_bin);

		double minimum = spreadComplexResult[j] * Constants.SORTSPREAD_MULTIPLIERS[0] + spreadComplexResult[j + 1]
				* Constants.SORTSPREAD_MULTIPLIERS[1] + spreadComplexResult[j + 2]
				* Constants.SORTSPREAD_MULTIPLIERS[2] + spreadComplexResult[j + 3]
				* Constants.SORTSPREAD_MULTIPLIERS[3];

		minimum = 20 * Math.log10(minimum * noiseThresholdShiftInAmplitude);
		return Math.min(minimum, spread_average);
	}

	public double postProcessFFTResults(FFTRecord fftRecord, int shiftbits) {
		double y = 0;
		final ComplexBuffer signal = fft.getX();
		for (int i = fftRecord.lo_bin; i <= fftRecord.hi_bin; i++) {
			final double x = signal.absolute(i) * skewing_function[i << shiftbits];
			spreadComplexResult[i] = x;
			y += x;
		}
		return 20 * Math.log10(y * fftRecord.oneOverNumbins * signalToNoiseRatioInAmplitude);
	}

	/**
	 * This operation obtains the minimum value in dB units for the significant
	 * bands. The algorithm first obtains the average amplitude for the bins
	 * inside the desired range, giving more importance to some bands vs others
	 * (skewing). Next, it calculates the minimum of those amplitudes, using
	 * some spreading from the contiguous bands. At last, the minimum of these
	 * two values is returned as the minimum.
	 * 
	 * @param currentFFTResult
	 *            FFT status variable
	 * @param analysisIndex
	 *            index of the analysis
	 * @return double with the minimum dB value of the analyzed bands.
	 */
	private double calculateMinimumdBNew(FFTRecord fftRecord, SpreadingRecord currentAnalysis, int shiftbits) {
		final ComplexBuffer signal = fft.getX();
		// Part one: Obtain the amplitude of the bins (the FFT result), for the
		// specific bins of this analysis.
		// Then, skew the signal to prefer some bands over others.
		// The operation is done in three steps, because we only care about the
		// amplitudes
		// between lo_bins and hi_bins.
		for (int idx = 0; idx <= fftRecord.hi_bin + 1; idx++) {
			spreadComplexResult[idx] = signal.absolute(idx) * skewing_function[idx << shiftbits];
		}

		double xAcc = 0.0;
		double spAcc = 0.0;
		double xMax = Constants.MAX_DB;
		double spMax = Constants.MAX_DB;
		for (int idx = fftRecord.lo_bin; idx <= fftRecord.max_bin; idx++) {
			final int width = currentAnalysis.spreadingAverage[idx].width;
			final int fraction = currentAnalysis.spreadingAverage[idx].fraction;

			double currentAcc = spreadComplexResult[idx];

			final double spread = ((spreadComplexResult[idx - 1] + spreadComplexResult[idx + 1])
					* spreadingNew[fraction].width + currentAcc)
					* spreadingNew[fraction].divisor;
			spAcc += spread;

			// Take the minimum of the RMS of the spreading.
			if (spMax > spread) {
				spMax = spread;
			}

			for (int j = 1; j < width; j++) {
				currentAcc += spreadComplexResult[idx + j];
			}
			currentAcc = currentAcc / width;
			xAcc += currentAcc;

			// Take the minimum of the RMS of the signal.
			if (xMax > currentAcc) {
				xMax = currentAcc;
			}
		}
		// Take the minimum between the RMS of the signal and the RMS of the
		// spreaded one.
		if (xMax > spMax) {
			xMax = spMax;
		}
		// Take the minimum of the accumulated spreaded signal and.... something
		// ?
		if (xAcc > spAcc) {
			xAcc = spAcc;
		}
		// Average the value for the accumulated one, and shift the max.
		xAcc = xAcc * fftRecord.oneOverNumbins * signalToNoiseRatioInAmplitude;
		xMax = xMax * noiseThresholdShiftInAmplitude;

		// And the winner is...
		if (xAcc > xMax) {
			xAcc = xMax;
		}
		// Convert the amplitude to dB
		return 20 * Math.log10(xAcc);
	}
}
