package org.lossywav.algorithm;
//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007, 2008 Nick Currie, Copyleft.
//    Copyright (C) 2008, C# port by Greg Chudov.
//    Copyright (C) 2008, Java port by Josep Maria Antoln
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> yahoo <dot> com
import java.io.IOException;
import java.io.PrintStream;
import java.util.Date;

import org.lossywav.audiocodecs.AudioDest;
import org.lossywav.audiocodecs.AudioSource;
import org.lossywav.audiocodecs.WavWriter;
import org.lossywav.util.CircularBuffer4;
import org.lossywav.util.Constants;

/**
 * Main lossyWav algorithm class.
 * @author JosepMa
 *
 */
public class LossyWav {
	// Input data
	private AudioSource audioSource;
	private AudioDest audioDest;
	private AudioDest correctionDest;
	//TODO:  parameter string should be generated from the setters of the parameters instead of rely in externally setting it.
	private String parameterString;

	//Algorithm related parameters
	private double quality = Constants.DEFAULT_QUALITY;
	private boolean forceFFT32 = false;
	private float frequencyLimit = Constants.DEFAULT_UPPER_FREQUENCY_LIMIT;
	private boolean linkChannelBitsToRemove = Constants.DEFAULT_LINK_CHANNELS;
	private int maxClipsAllowed = Constants.MIN_CLIPS-1;
	private double shapingFactor = Constants.MIN_SHAPING-1;
	private boolean useMidSide = Constants.DEFAULT_MID_SIDE;
	private boolean sortSpread = Constants.DEFAULT_SORTSPREAD;
	private int underlap = Constants.DEFAULT_UNDERLAP;
	//TODO: Use it.
	private boolean checkIfProcessed = false;
	private double deltaNTS = 0.0;
	private double deltaSNR = 0.0;

	//Internal parameters
	private transient CircularBuffer4 readBlocks;
	private transient final Analyzer analyzer = new Analyzer();
	private transient final LossyWavStatus status = new LossyWavStatus();
	private transient ChannelStatus[] channelStatus;
	private transient double midChannelRMSBits;
	private transient double sideChannelRMSBits;


	public LossyWav() {
		super();
	}
	/**
	 * getter of audioSource
	 * @return the audioSource
	 */
	public AudioSource getAudioSource() {
		return audioSource;
	}
	/**
	 * setter of audioSource
	 * @param audioSource the audioSource to set
	 */
	public void setAudioSource(AudioSource audioSource) {
		this.audioSource = audioSource;
	}
	/**
	 * getter of audioDest
	 * @return the audioDest
	 */
	public AudioDest getAudioDest() {
		return audioDest;
	}
	/**
	 * setter of audioDest
	 * @param audioDest the audioDest to set
	 */
	public void setAudioDest(AudioDest audioDest) {
		this.audioDest = audioDest;
	}
	/**
	 * getter of correctionDest
	 * @return the correctionDest
	 */
	public AudioDest getCorrectionDest() {
		return correctionDest;
	}
	/**
	 * setter of correctionDest
	 * @param correctionDest the correctionDest to set
	 */
	public void setCorrectionDest(AudioDest correctionDest) {
		this.correctionDest = correctionDest;
	}
	public String getParameterString() {
		return parameterString;
	}
	public void setParameterString(String parameterString) {
		this.parameterString = parameterString;
	}

	/**
	 * getter of quality
	 * @return quality
	 */
	public double getQuality() {
		return this.quality;
	}
	/**
	 * setter of quality
	 * @param quality double
	 */
	public void setQuality(double quality) {
		this.quality = quality;
		if (status.isInitialized()) {
			setQualityInternal(quality);
		}
	}
	public int getAnalysisNumber() {
		return analyzer.getAnalysisMode();
	}
	public void setAnalysisNumber(int analysisNumber) {
		analyzer.setAnalysisMode(analysisNumber);
	}
	public float getFrequencyLimit() {
		return frequencyLimit;
	}
	public void setFrequencyLimit(float frequencyLimit) {
		this.frequencyLimit = frequencyLimit;
	}	
	public boolean isLinkChannelBitsToRemove() {
		return linkChannelBitsToRemove;
	}
	public void setLinkChannelBitsToRemove(boolean linkChannelBitsToRemove) {
		this.linkChannelBitsToRemove = linkChannelBitsToRemove;
	}	
	/**
	 * getter of maxClipsAllowed
	 * @return the maxClipsAllowed
	 */
	public int getMaxClipsAllowed() {
		return maxClipsAllowed;
	}
	/**
	 * setter of maxClipsAllowed
	 * @param maxClipsAllowed the maxClipsAllowed to set
	 */
	public void setMaxClipsAllowed(int maxClipsAllowed) {
		this.maxClipsAllowed = maxClipsAllowed;
		if (status.isInitialized()) {
			setQualityInternal(quality);
		}
	}
	/**
	 * getter of useMidSide
	 * @return the useMidSide
	 */
	public boolean isUseMidSide() {
		return useMidSide;
	}
	/**
	 * setter of useMidSide
	 * @param useMidSide the useMidSide to set
	 */
	public void setUseMidSide(boolean useMidSide) {
		this.useMidSide = useMidSide;
	}
	public double getScalingFactor() {
		return analyzer.getScalingFactor();
	}
	public void setScalingFactor(double scalingFactor) {
		analyzer.setScalingFactor(scalingFactor);
	}
	public double getShapingFactor() {
		return shapingFactor;
	}
	public void setShapingFactor(double shapingFactor) {
		this.shapingFactor = shapingFactor;
		if (status.isInitialized()) {
			setQualityInternal(quality);
		}
	}
	public boolean isForceFFT32() {
		return forceFFT32;
	}
	public void setForceFFT32(boolean forceFFT32){
		this.forceFFT32 = forceFFT32;
	}	
	/**
	 * getter of sortSpread
	 * @return the sortSpread
	 */
	public boolean isSortSpread() {
		return sortSpread;
	}
	/**
	 * setter of sortSpread
	 * @param sortSpread the sortSpread to set
	 */
	public void setSortSpread(boolean sortSpread) {
		this.sortSpread = sortSpread;
	}
	/**
	 * getter of underlap
	 * @return the underlap
	 */
	public int getUnderlap() {
		return underlap;
	}
	/**
	 * setter of underlap
	 * @param underlap the underlap to set
	 */
	public void setUnderlap(int underlap) {
		this.underlap = underlap;
	}

	/**
	 * getter of checkIfProcessed
	 * @return the checkIfProcessed
	 */
	public boolean isCheckIfProcessed() {
		return checkIfProcessed;
	}
	/**
	 * setter of checkIfProcessed
	 * @param checkIfProcessed the checkIfProcessed to set
	 */
	public void setCheckIfProcessed(boolean checkIfProcessed) {
		this.checkIfProcessed = checkIfProcessed;
	}
	
	/**
	 * getter of deltaNTS
	 * @return the deltaNTS
	 */
	public double getDeltaNTS() {
		return deltaNTS;
	}
	/**
	 * setter of deltaNTS
	 * @param deltaNTS the deltaNTS to set
	 */
	public void setDeltaNTS(double deltaNTS) {
		this.deltaNTS = deltaNTS;
	}
	/**
	 * getter of deltaSNR
	 * @return the deltaSNR
	 */
	public double getDeltaSNR() {
		return deltaSNR;
	}
	/**
	 * setter of deltaSNR
	 * @param deltaSNR the deltaSNR to set
	 */
	public void setDeltaSNR(double deltaSNR) {
		this.deltaSNR = deltaSNR;
	}
	public LossyWavStatus getStatus() {
		return status;
	}

	public void process(PrintStream out) throws IOException
	{
		int sampleRate = audioSource.getSampleRate();
		if (this.frequencyLimit > sampleRate*0.5f) {
			frequencyLimit = sampleRate*0.5f;
		}
		if (audioSource.getChannelCount() != 2 && useMidSide) {
			useMidSide = false;
		}
		int codecBlockSize;
		if (sampleRate < 69075) {
			codecBlockSize = 512;
		}
		else if ( sampleRate < 138150 ) {
			codecBlockSize = 1024;
		}
		else if ( sampleRate < 276300 ) {
			codecBlockSize = 2048;
		}
		else {
			codecBlockSize = 4096;
		}
		if (!status.isInitialized()) {
			initialize(codecBlockSize);
		}
		writeLossyWavTag();
        out.println(String.format("Filename  : %s", audioSource.getPath()));
        out.println(String.format("File Info : %.02fkHz; %d channel; %d bit; %d samples;", 
        		audioSource.getSampleRate()*0.001,audioSource.getChannelCount(),audioSource.getBitsPerSample(),
        		audioSource.getLength()));

		long i = 0;
		long printAmount =  audioSource.getLength() / (codecBlockSize * 10);
		int nextSize = readNextCodecBlock(codecBlockSize);
		while (nextSize > 0)
		{
			readBlocks.rotate();
			calculateSignalRMSInBits(readBlocks.getCurrentBuffer(), readBlocks.getCurrentLength());
			// Read one buffer ahead.
			final int currentSize = nextSize;
			nextSize = readNextCodecBlock(codecBlockSize);
			processCurrentCodecBlock();
			if (audioDest != null) {
				audioDest.write(analyzer.getResultBlock(), currentSize);
			}
			if (correctionDest != null) {
				correctionDest.write(analyzer.getCorrectionBlock(), currentSize);
			}
			if (i++ >= printAmount) {
				out.format("Position: %.2f%%\n" , 100.0 * audioSource.getPosition()/(float)audioSource.getLength());
				i=0;
			}
		}
		out.format("Position: 100%%\n");
	}
	public void merge(PrintStream out) throws IOException {
		
	}
	
	private void setQualityInternal(double newQual) {
		analyzer.setFFT32( forceFFT32 || Constants.QUALITY_AUTO_FFT32_ON[(int) Math.floor(newQual - Constants.MIN_QUALITY)] );

		if ( shapingFactor >= Constants.MIN_SHAPING) {
			analyzer.setShapingFactor(shapingFactor);
		} else {
			analyzer.setShapingFactor(Math.max(0.0, newQual / 10.0));
		}

		analyzer.setSortSpread(sortSpread);
		double noiseThreshold = 0.0f;
		if (sortSpread) {
			noiseThreshold = interpolateParam(Constants.QUALITY_NOISE_THRESHOLDS_SHIFT, newQual) + interpolateParam(Constants.QUALITY_NOISE_THRESHOLDS_SHIFT_DELTA, newQual);
		} else {
			noiseThreshold = interpolateParam(Constants.QUALITY_NOISE_THRESHOLDS_SHIFT, newQual);
		}
		if (deltaNTS != 0.0) {
			noiseThreshold+=deltaNTS;
		}
		analyzer.setNoiseThresholdShift( Math.round(noiseThreshold * 1000) / 1000.0f );

		double signalToNoise =interpolateParam(Constants.QUALITY_SIGNAL_TO_NOISE_RATIO, newQual);
		if (deltaSNR != 0.0) {
			signalToNoise += deltaSNR; 
		}
		analyzer.setSnrValue( Math.round( signalToNoise * 1000) / 1000.0f );
		analyzer.setQualityBasedMinimumBitsToKeep( Math.round(interpolateParam(Constants.QUALITY_DYNAMIC_MINIMUM_BITS_TO_KEEP, newQual) * 1000) / 1000.0f );
		if ( maxClipsAllowed >= Constants.MIN_CLIPS) {
			analyzer.setMaximumClipsPerChannel( maxClipsAllowed );	
		} else {
			analyzer.setMaximumClipsPerChannel( (int)Math.round(interpolateParam(Constants.QUALITY_MAXIMUM_CLIPS_PER_CHANNEL, newQual)) );
		}
	}

	private double interpolateParam(double[] param, double quality)
	{
		if (quality >= Constants.MAX_QUALITY) {
			return param[(int) (Constants.MAX_QUALITY - Constants.MIN_QUALITY)];
		}
		final double qualityInArray = quality - Constants.MIN_QUALITY;
		final int qualityInteger = (int) Math.floor(qualityInArray);
		final double qualityFraction = qualityInArray - qualityInteger;
		return qualityFraction * (param[qualityInteger + 1] - param[qualityInteger]) + param[qualityInteger];
	}

	private void initialize(int codecBlockLength) {
		readBlocks = new CircularBuffer4( codecBlockLength, audioSource.getChannelCount());

		channelStatus = new ChannelStatus[audioSource.getChannelCount()];
		for (int i=0; i < audioSource.getChannelCount(); i++) {
			channelStatus[i] =  new ChannelStatus();
		}

		setQualityInternal(quality);
		analyzer.initialize(audioSource.getChannelCount(), audioSource.getBitsPerSample(), 
				audioSource.getSampleRate(), codecBlockLength, frequencyLimit, underlap);

		status.reset();
		status.setInitialized(true);
	}

	private void writeLossyWavTag() throws IOException  {
		final String version_string = Constants.LOSSYWAV_TAG + "/" + Constants.JLOSSYWAV_TAG + "#";
		final String datestamp = new Date().toString();
		final String factString = version_string+" @ "+datestamp+", "+parameterString + "\r\n\0";
		if (audioDest != null && audioDest instanceof WavWriter) {
			((WavWriter) audioDest).WriteChunk(WavWriter.fccFact, factString.getBytes());
		}
		if (correctionDest != null && correctionDest instanceof WavWriter){
			((WavWriter) correctionDest).WriteChunk(WavWriter.fccFact, factString.getBytes());
		}
	}
	private int readNextCodecBlock(int codecBlockSize) throws IOException	{
		final int size = Math.min((int)audioSource.getRemaining(), codecBlockSize);
		if ( size > 0) {
			final int readSize = audioSource.read(readBlocks.getNextBuffer(), size);
			if (readSize != size) {
				throw new IOException("Problem reading from input. Expected bytes: " + size + ", got: " + readSize);
			}
			readBlocks.setNextLength(readSize);
		}
		return size;
	}

	private void processCurrentCodecBlock() {
		final short channels = (short) audioSource.getChannelCount();
		int bitsToRemoveThisBlock = audioSource.getBitsPerSample();

		if (linkChannelBitsToRemove) {
			double minCurrentBlockBits = channelStatus[0].currentBits;
			for (int channel = 1; channel < channels; channel++){
				minCurrentBlockBits = Math.min(minCurrentBlockBits, channelStatus[channel].currentBits);
			}
			if (useMidSide) {
				minCurrentBlockBits = Math.min(minCurrentBlockBits, midChannelRMSBits);
				minCurrentBlockBits = Math.min(minCurrentBlockBits, sideChannelRMSBits);
			}
			for (int channel = 0; channel < channels; channel++) {
				channelStatus[channel].currentBits = minCurrentBlockBits;
			}
		}

		for (int channel = 0; channel < channels; channel++) {
			analyzer.process(channel, readBlocks, channelStatus[channel], status);
			bitsToRemoveThisBlock = Math.min(bitsToRemoveThisBlock,
					channelStatus[channel].bitsToRemove);
		}

		for (int channel = 0; channel < channels; channel++)
		{
		   /* TODO:
		    * if channels_linked then
		      begin
		        bits_to_remove_from_this_channel:=codec_block_dependent_bits_to_remove;
		        if bits_to_remove_channel[channel]<>bits_to_remove_from_this_channel then
		        begin
		          if ( channel < channels ) {
		            // prepare FIFO buffer with saved state information for this channel;
		            for ( int j=0; j < Constants.SHAPING_LENGTH ; j++ ) {
				          fifo[j] = saved_fifo_state[channel][j];
				    }
		            remove_bits(channel, minimumFFTResult.btr, channelStatus, rotatingBlocks);
		            // save FIFO state information to temp storage
			        for ( int j = 0; j < Constants.SHAPING_LENGTH; j++ ) {
			          temp_fifo_state[channel][j] = fifo[j + codecBlockSize];
			        }
		          }
		          bits_to_remove_channel[channel]:=bits_to_remove_from_this_channel;
		
		          bits_lost_this_channel[channel]:=min_fft_result.btr-bits_to_remove_from_this_channel;
		          clipped_samples_this_channel[channel]:=this_channel_clips;
		        end;
		      end;
			}*/
			status.addOverallBitsRemoved(channelStatus[channel].bitsToRemove);
			status.addOverallBitsLost(channelStatus[channel].bitsLost);
			status.addClippedSamples(channelStatus[channel].clippedSamples);
		}
		/*
	    for (int channel = 0; channel < channels; channel++) {
	          // copy FIFO saved state from temp storage to saved storage, i.e. no more bit-removal for this codec-block.
	        for ( int j = 0; j < Constants.SHAPING_LENGTH; j++ ) {
	          saved_fifo_state[channel][j]:=temp_fifo_state[channel][j];
	         }
		}
	    samples_processed:=(blocks_processed-1)*codec_block_size+nwav_this_block_size;
		 */
	}

	private void calculateSignalRMSInBits(int[][] block, int length) {
		status.increaseProcessedBlocks(1);
		final double logLength = Math.log10(length);
		for (int channel = 0; channel < audioSource.getChannelCount(); channel++){
			double x = 0;
			for (int i = 0; i < length; i++)
			{
				final double s = block[channel][i];
				x += s * s;
			}
			//channelStatus[channel].currentRms = 10.0 * (Math.log10(x * 4) - logLength); // * 4 is equivalent of adding 6.02dB's.
			 channelStatus[channel].currentBits = (Math.log10(x) - logLength) * Constants.INVLOG2X2;
		}
		
		if (useMidSide) {
			int x = 0;
			int y = 0;
			for (int i=0; i < length; i++ ) {
				final double right = block[1][i];
				final double mid = 0.5 * (block[0][i]+right);
				final double side = mid-right;
				// TODO:
				// Here the LossyWav implementation puts
				// mid and side into blocks 2 and 3
				x += mid * mid;
				y += side * side;
			}
			this.midChannelRMSBits = (Math.log10(x) - logLength) * Constants.INVLOG2X2;;
			this.sideChannelRMSBits = (Math.log10(y) - logLength) * Constants.INVLOG2X2;;
		} else {
			this.midChannelRMSBits = 0.0;
			this.sideChannelRMSBits = 0.0;
		}
	}
}
