package org.lossywav.audiocodecs;
//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007, 2008 Nick Currie, Copyleft.
//    Copyright (C) 2008, C# port by Greg Chudov.
//    Copyright (C) 2008, Java port by Josep Maria Antoln
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> yahoo <dot> com
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.Map;

import org.lossywav.util.Constants;
/**
 * Wav Reader implementation of AudioSource
 * @author JosepMa
 *
 */
public class WavReader implements AudioSource {
	/**
	 * RIFF tag.
	 */
	private static final String RIFF = "RIFF";
	/**
	 * Wave tag.
	 */
	private static final String WAVE = "WAVE";
	/**
	 * fmt tag.
	 */
	private static final String FORMAT = "fmt ";
	/**
	 * PCM (integer) format value.
	 */
	private static final short WAVEFMT_PCM = 1;
	/**
	 * Data tag.
	 */
	private static final String DATA = "data";
	/**
	 * Position in bytes of the first sample inside the file.
	 */
	private long dataPosition;
	/**
	 * position, in samples, read fo far.
	 */
	private long position;
	/**
	 * length in samples of this audio source.
	 */
	private long length;
	/**
	 * bitswidth of a sample of a single channel.
	 */
	private short bitsPerSample;
	/**
	 * number of channels.
	 */
	private short channelCount;
	/**
	 * sampling rate of the samples.
	 */
	private int sampleRate;
	/**
	 * amount of bytes of a complete sample ((bytesPerSamples + bytePadding) * channelcount).
	 */
	private short blockAlign;
	/**
	 * String with the path to the file being read.
	 */
	private final String path;
	/**
	 * File class element.
	 */
	private transient final RandomAccessFile file;
	/**
	 * Buffer used for reads and conversion to little endian.
	 */
	private transient final ByteBuffer wavBuffer;
	/**
	 * Wave Reader constructor
	 * @param path path to the file to be read.
	 * @throws IOException
	 */
	public WavReader(String path) throws IOException
	{
		this.path = path;
		this.file = new RandomAccessFile(path, "r");
		readHeaders();
		// Set the buffer size to an exact multiplier of blockAlign
		wavBuffer = ByteBuffer.allocate((Constants.BUFFER_SIZE / blockAlign) * blockAlign);
		wavBuffer.order(ByteOrder.LITTLE_ENDIAN);
	}
	/**
	 * Get the bitwidth of each sample of this audio source.
	 * @return int
	 */
	public int getBitsPerSample() {
		return bitsPerSample;
	}
	/**
	 * Get the number of audio channels of this audio source.
	 * @return int
	 */
	public int getChannelCount() {
		return channelCount;
	}
	/**
	 * get the block alignment of this source.
	 *  ((bytesPerSamples + bytePadding) * channelcount).
	 * @return
	 */
	public int getBlockAlign() {
		return blockAlign;
	}
	/**
	 * Get the length in samples of this audio source, if applicable.
	 * @return int
	 */
	public long getLength() {
		if (length == 0) {
			try {
				findDataTag();
			} catch (final IOException e) {
				e.printStackTrace();
			}
		}
		return length;
	}
	/**
	 * Get the path or description of this input element.
	 * @return String
	 */
	public String getPath() {
		return path;
	}
	/**
	 * Get the position in samples of the cursor.
	 * @return long
	 */
	public long getPosition() {
		return position;
	}
	/**
	 * Set the position in samples of the cursor.
	 * @param position long
	 * @throws IOException
	 */
	public void setPosition(long position) throws IOException {
		long seekPos;
		long curPos = getPosition();

		if (position > length) {
			this.position = this.length;
		}
		else {
			this.position = position;
		}
		
		int difference = (int)(this.position - curPos) * blockAlign;
		seekPos = dataPosition + this.position * blockAlign;
		if ( difference < 0 || difference > wavBuffer.remaining()) {
			file.seek(seekPos);
			file.read(wavBuffer.array(), 0, Math.min(wavBuffer.array().length, (int) getRemaining() * blockAlign));
			wavBuffer.rewind();
		}
		else {
			wavBuffer.position(wavBuffer.position() + difference);
		}
	}
	/**
	 * Get the amount of samples remaining to be read ( length - position ).
	 * @return long
	 */
	public long getRemaining() {
		if (length == 0) {
			try {
				findDataTag();
			} catch (final IOException e) {
				e.printStackTrace();
			}
		}
		return length - position;
	}
	/**
	 * Get the sampling rate of the samples of this audio source.
	 * @return int
	 */
	public int getSampleRate() {
		return sampleRate;
	}
	/**
	 * Obtain the metadata tags of this audio source.
	 * @return tags Map of tag name and tag content.
	 */
	public Map<String, String> getTags() {
		// TODO Auto-generated method stub
		return null;
	}
	/**
	 * Write metadata (tags) to the output.
	 * @param tags Map of tag name and tag content.
	 */
	public void setTags(Map<String, String> tags) {
		// TODO Auto-generated method stub

	}
	/**
	 * Close the output, writing any data pending, and free any resources used.
	 * @throws IOException
	 */
	public void close() throws IOException {
		file.close();

	}
	/**
	 * Read samples from the audio source.
	 * @param buffer array of [channel][samples] with the samples
	 * @param sampleCount amount of samples to get.
	 * @return the real amount of samples read. Should be the same than sampleCount if everything is ok
	 * @throws IOException
	 */
	//TODO: This process wouldn't work properly when blockAlign is not exacty bitsPerSample / 8
	// (i.e. 24bits aligned to 32 and 12 or 20bits)
	public int read(int[][] buffer, int sampleCount)  throws IOException{
		if (length == 0) {
			findDataTag();
		}
		final int bytesPerSample = blockAlign / channelCount;
		int samplesToRead =  Math.min(sampleCount, (wavBuffer.remaining()) / blockAlign);
		int samplesTotal = 0;

		while (samplesTotal < sampleCount) {
			switch(bytesPerSample) {
			case 1:
				for (int i=0; i < samplesToRead; i++) {
					for (int c=0; c < channelCount; c++) {
						buffer[c][i] = wavBuffer.get();
					}
				}
				break;
			case 2:
				for (int i=0; i < samplesToRead; i++) {
					for (int c=0; c < channelCount; c++) {
						buffer[c][i] = wavBuffer.getShort();
					}
				}
				break;
			case 3:
				for (int i=0; i < samplesToRead; i++) {
					for (int c=0; c < channelCount; c++) {
						final short lobyte = wavBuffer.getShort();
						final byte hibyte = wavBuffer.get();
						buffer[c][i] = hibyte << 16 + lobyte;
					}
				}
			break;
			case 4:
				for (int i=0; i < samplesToRead; i++) {
					for (int c=0; c < channelCount; c++) {
						buffer[c][i] = wavBuffer.getInt();
					}
				}
				break;
			default:
				throw new IOException("Unexpected sample size:" + bytesPerSample);
			}
	
			position += samplesToRead;
			samplesTotal += samplesToRead;
			if ( wavBuffer.remaining() == 0 ) {
				file.read(wavBuffer.array(), 0, Math.min(wavBuffer.array().length, (int) getRemaining() * blockAlign));
				wavBuffer.rewind();
				samplesToRead =  Math.min(sampleCount - samplesTotal, wavBuffer.array().length / blockAlign);
			}
		}
		return samplesTotal;
	}
	/**
	 * Locates and reads a chunck named like the string passed.
	 * @param fcc String with the name of the chunk
	 * @return array of bytes with the content
	 * @throws IOException
	 */
	public byte[] readChunk(String fcc) throws IOException
	{
		final long position = file.getFilePointer();
		String tag = readTag();
		while( tag != null && !tag.equals(fcc)) {
			file.skipBytes(readLEInt());
			tag = readTag();
		}

		if (tag == null) {
			file.seek(position);
			throw new IOException("couldn't locate the tag:" + fcc);
		}
		final int chunkSize= readLEInt();
		final byte[] result = new byte[chunkSize];
		file.read(result);
		file.seek(position);
		return result;
	}
	/**
	 * reads the headers of the file.
	 * @throws IOException
	 */
	private void readHeaders() throws IOException {

        String tag = readTag();
		if (!tag.equals(RIFF)) {
			throw new IOException("Unexepected type of file:" + tag.toString());
		}
		file.skipBytes(4);
		tag = readTag();
		if (!tag.equals(WAVE)) {
			throw new IOException("Unexepected type of file:" + tag.toString());
		}
		tag = readTag();
		if (!tag.equals(FORMAT)) {
			throw new IOException("Unexepected type of file:" + tag.toString());
		}
		final int chunkSize = readLEInt();
		final short format = readLEShort();
		if (format != WAVEFMT_PCM) {
			throw new IOException("Unsupported Wave Format:" + format);
		}
		channelCount = readLEShort();
		sampleRate = readLEInt();
		file.skipBytes(4); // bytes per second
		blockAlign = readLEShort();
		bitsPerSample = readLEShort();

		file.skipBytes(chunkSize-16);
	}
	/**
	 * Locates and puts itself in the data tag, to start reading the samples.
	 * @throws IOException
	 */
	private void findDataTag() throws IOException {
		final long curPosition = file.getFilePointer();
		String tag = readTag();
		while( tag != null && !tag.equals(DATA)) {
			file.skipBytes(readLEInt());
			tag = readTag();
		}

		if (tag == null) {
			file.seek(curPosition);
			throw new IOException("couldn't locate the tag:" + DATA);
		}
		int lengthBytes = readLEInt();
		length = lengthBytes / blockAlign;
		position = 0;
		dataPosition = file.getFilePointer();
		file.read(wavBuffer.array(), 0, Math.min(wavBuffer.array().length, lengthBytes));
		wavBuffer.rewind();
	}
	/**
	 * Read a tag String from the file.
	 * @return String with the tag
	 * @throws IOException
	 */
	private String readTag() throws IOException {
		final char[] tagChar = new char[4];
		for(int i=0; i < 4; i++) {
			tagChar[i] = (char)  file.readUnsignedByte();
		}
		return String.valueOf(tagChar);
	}
	/**
	 * Read an int with Litte Endian.
	 * @return int
	 * @throws IOException
	 */
	private int readLEInt() throws IOException {
		return Integer.reverseBytes(file.readInt());
	}
	/**
	 * Read a short with Little Endian.
	 * @return short
	 * @throws IOException
	 */
	private short readLEShort() throws IOException {
		return Short.reverseBytes(file.readShort());
	}
}
