package org.lossywav.audiocodecs;
//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007, 2008 Nick Currie, Copyleft.
//    Copyright (C) 2008, C# port by Greg Chudov.
//    Copyright (C) 2008, Java port by Josep Maria Antoln
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> yahoo <dot> com
import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.Map;

import org.lossywav.util.Constants;

public class WavWriter implements AudioDest {
	/**
	 * ffc for the FACT chunk.
	 */
	public static final String fccFact ="fact";
	/**
	 * RIFF tag.
	 */
	private static final String RIFF = "RIFF";
	/**
	 * Wave tag.
	 */
	private static final String WAVE = "WAVE";
	/**
	 * fmt tag.
	 */
	private static final String FORMAT = "fmt ";
	/**
	 * PCM (integer) format value.
	 */
	private static final short WAVEFMT_PCM = 1;
	/**
	 * Size in bytes of the PCM (integer) format chunk.
	 */
	private static final int WAVEFMT_PCM_SIZE = 16;
	/**
	 * Data tag.
	 */
	private static final String DATA = "data";
	/**
	 * Max writeable file size.
	 */
	private static final int MAX_FILE_SIZE = 0x7FFFFFFE;
	/**
	 * length in samples of this audio source.
	 */
	private long length;
	/**
	 * bitswidth of a sample of a single channel.
	 */
	private final short bitsPerSample;
	/**
	 * number of channels.
	 */
	private final short channelCount;
	/**
	 * sampling rate of the samples.
	 */
	private final int sampleRate;
	/**
	 * amount of bytes of a complete sample ((bytesPerSamples + bytePadding) * channelcount).
	 */
	private final short blockAlign;
	/**
	 * String with the path to the file being read.
	 */
	private final String path;
	/**
	 * File class element.
	 */
	private transient final RandomAccessFile file;
	/**
	 * size in bytes of the header.
	 */
	private int hdrLen;
	/**
	 * Buffer used for writes and conversion to little endian.
	 */
	private transient final ByteBuffer wavBuffer;	
	/**
	 * WavWriter constructor.
	 * @param path String with the path of the file
	 * @param bitsPerSample the bits per sample that it has
	 * @param channelCount the number of channels that it has
	 * @param sampleRate the sampling rate of the samples
	 */
	public WavWriter(String path, short bitsPerSample, short channelCount,
			int sampleRate) throws IOException {

		this.path = path;
		this.bitsPerSample = bitsPerSample;
		this.channelCount = channelCount;
		this.sampleRate = sampleRate;
		this.blockAlign = (short)(channelCount * (short)((bitsPerSample + 7) / 8));
		this.file = new RandomAccessFile(path, "rw");
		// Set the buffer size to an exact multiplier of blockAlign
		wavBuffer = ByteBuffer.allocate((Constants.BUFFER_SIZE / blockAlign) * blockAlign);
		wavBuffer.order(ByteOrder.LITTLE_ENDIAN);
		writeHeaders();
	}
	/**
	 * Get the path or description of this output element.
	 * @return String
	 */
	public String getPath() {
		return path;
	}
	/**
	 * Write metadata (tags) to the output.
	 * @param tags Map of tag name and tag content.
	 */
	public void setTags(Map<String, String> tags) {
		// TODO Auto-generated method stub

	}
	/**
	 * Close the output, writing any data pending, and free any resources used.
	 * @throws IOException
	 */
	public void close() throws IOException {

		long dataLen, dataLenPadded;

		dataLen = length * blockAlign;
		if (wavBuffer.position() != 0) {
			file.write(wavBuffer.array(), 0, wavBuffer.position());
		}

		if (dataLen + hdrLen > MAX_FILE_SIZE) {
			dataLen = (MAX_FILE_SIZE - hdrLen) / blockAlign * blockAlign;
		}
		//Requires padding?
		if ((dataLen & 1) > 0) {
			file.writeByte(0);
		}

		dataLenPadded = dataLen + (dataLen & 1);

		file.seek(4);
		writeLEInt((int) (dataLenPadded + hdrLen - 8));

		file.seek(hdrLen-4);
		writeLEInt((int) dataLen);

		file.close();
	}
	/**
	 * Close and delete the file from the disk, if applicable.
	 * @throws IOException
	 */
	public void delete() throws IOException {
		file.close();

		final File physFile = new File(path);
		if (physFile.exists()) {
			if (!physFile.delete()){
				throw new IOException("file.delete call failed");
			}
		}
	}
	/**
	 * Write audio data to the output.
	 * @param buffer array with [channel][sample] of type int with the sample values
	 * @param sampleCount int amount of samples to read from the buffer
	 * @throws IOException any problem related to writing to the file
	 */
	public void write(int[][] buffer, int sampleCount) throws IOException {
		final int bytesPerSample = blockAlign / channelCount;
		
		int samplesToWrite =  Math.min(sampleCount, wavBuffer.remaining()/ blockAlign);
		int samplesTotal = 0;

		while (samplesTotal < sampleCount) {

			switch(bytesPerSample) {
			case 1:
				for (int i=0; i < samplesToWrite; i++) {
					for (int c=0; c < channelCount; c++) {
						wavBuffer.put((byte) buffer[c][i]);
					}
				}
				break;
			case 2:
				for (int i=0; i < samplesToWrite; i++) {
					for (int c=0; c < channelCount; c++) {
						wavBuffer.putShort((short) buffer[c][i]);
					}
				}
				break;
			case 3:
				for (int i=0; i < samplesToWrite; i++) {
					for (int c=0; c < channelCount; c++) {
						wavBuffer.putShort((short) buffer[c][i]);
						final byte hibyte = (byte) (buffer[c][i] >> 16);
						wavBuffer.put(hibyte);
					}
				}
			break;
			case 4:
				for (int i=0; i < samplesToWrite; i++) {
					for (int c=0; c < channelCount; c++) {
						wavBuffer.putInt(buffer[c][i]);
					}
				}
				break;
			default:
				throw new IOException("Unexpected sample size:" + bytesPerSample);
			}
			length += samplesToWrite;
			samplesTotal += samplesToWrite;
			if ( wavBuffer.remaining() == 0 ) {
				file.write(wavBuffer.array());
				wavBuffer.rewind();
				samplesToWrite =  Math.min(sampleCount - samplesTotal, wavBuffer.array().length / blockAlign);
			}			
		}
	}
	/**
	 * writes a chunk of data named after the string in fcc.
	 * Warning: Chunks always go before the DATA chunk. Do not write DATA before
	 * adding a chunk.
	 * @param fcc String with the chunk name
	 * @param data byte array with the data to be stored
	 * @throws IOException
	 */
	public void WriteChunk(String fcc, byte[] data) throws IOException
	{
		if (length > 0) {
			throw new IOException("data already written, no chunks allowed");
		}
		int dataLen = data.length;

		file.seek(hdrLen - 8);
		file.writeBytes(fcc);
		writeLEInt(data.length);
		file.write(data);
		//Requires padding?
		if ((data.length & 1) > 0) {
			file.writeByte(0);
			dataLen++;
		}
		file.writeBytes(DATA);
		writeLEInt(0);
		hdrLen += dataLen + 8;
	}
	/**
	 * Writes the headers of this file.
	 * @throws IOException
	 */
	private void writeHeaders() throws IOException
	{
		file.setLength(0);
		file.writeBytes(RIFF);
		writeLEInt(0);
		file.writeBytes(WAVE);

		file.writeBytes(FORMAT);
		writeLEInt(WAVEFMT_PCM_SIZE);
		writeLEShort(WAVEFMT_PCM);
		writeLEShort(channelCount);
		writeLEInt(sampleRate);
		writeLEInt(sampleRate * blockAlign);
		writeLEShort(blockAlign);
		writeLEShort(bitsPerSample);

		file.writeBytes(DATA);
		writeLEInt(0);
		hdrLen = 44;
	}
	/**
	 * Write an int with Little Endian.
	 * @param value int
	 * @throws IOException
	 */
	private void writeLEInt(int value) throws IOException {
		file.writeInt(Integer.reverseBytes(value));
	}
	/**
	 * Write a short with Little Endian.
	 * @param short value int
	 * @throws IOException
	 */
	private void writeLEShort(short value) throws IOException {
		file.writeShort(Short.reverseBytes(value));
	}
}
