package org.lossywav.dsp;
//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007, 2008 Nick Currie, Copyleft.
//    Copyright (C) 2008, C# port by Greg Chudov.
//    Copyright (C) 2008, Java port by Josep Maria Antoln
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> yahoo <dot> com
/**
 * Interface for a Complex Buffer.
 * @author JosepMa
 *
 */
public interface ComplexBuffer {
	/**
	 * Returns the value of a position as a Complex value.
	 * @param position position index in the buffer
	 * @return Complex with the value.
	 */
	Complex getComplex(int position);
	/**
	 * Sets a position to a new values.
	 * @param position position index in the buffer
	 * @param complex complex value
	 */
	void setComplex(int position, Complex complex);
	/**
	 * Sets a position to a new value, using double values.
	 * @param position position index in the buffer
	 * @param real real part of the complex number
	 * @param img imaginary part of the complex number
	 */
	void setComplex(int position, double real, double img);
	/**
	 * Getter for the imaginary part of a value.
	 * @param position position index in the buffer
	 * @return double value with the imaginary part
	 */
	double getImag(int position);
	/**
	 * Setter for the imaginary part of a value.
	 * @param position position index in the buffer
	 * @param imag double with the imaginary value.
	 */
	void setImag(int position, double imag);
	/**
	 * getter for the real part of a value.
	 * @param position position index in the buffer
	 * @return double with the real part
	 */
	double getReal(int position);
	/**
	 * setter for the real part of a value.
	 * @param position position index in the buffer
	 * @param real with the real value
	 */
	void setReal(int position, double real);
	/**
	 * sums the values of two positions and returns the result as a Complex.
	 * @param positionA position index in the buffer for the first element
	 * @param positionB position index in the buffer for the second element
	 * @return Complex with the value of the operation
	 */
	Complex sumPositions(int positionA, int positionB);
	/**
	 * sums the values of two positions and stores the value in positionA.
	 * @param positionA position index in the buffer for the first element
	 * @param positionB position index in the buffer for the second element
	 * @return Complex with the value of the operation
	 */
	void sumPositionsInA(int positionA, int positionB);
	/**
	 * subtracts the values of two positions and returns the result as a Complex.
	 * @param positionA position index in the buffer for the first element
	 * @param positionB position index in the buffer for the second element
	 */
	Complex substractPositions(int positionA, int positionB);
	/**
	 * subtracts the values of two positions and stores the value in positionA.
	 * @param positionA position index in the buffer for the first element
	 * @param positionB position index in the buffer for the second element
	 * @return Complex with the value of the operation
	 */
	void substractPositionsInA(int positionA, int positionB);
	/**
	 * multiplies the values of two positions and returns the result as a Complex.
	 * @param positionA position index in the buffer for the first element
	 * @param positionB position index in the buffer for the second element
	 */
	Complex multiplyPositions(int positionA, int positionB);
	/**
	 * multiplies the values of two positions and stores the value in positionA.
	 * @param positionA position index in the buffer for the first element
	 * @param positionB position index in the buffer for the second element
	 * @return Complex with the value of the operation
	 */
	void multiplyPositionsInA(int positionA, int positionB);
	/**
	 * Multiplies complexA with the complex in positionB and returns the real part of that operation.
	 * @param complexA complex number to operate
	 * @param positionB index of the complex number in the array
	 */
	double multiplyComplexWithPositionAndReturnReal(Complex complexA, int positionB);
	/**
	 * Multiplies complexA with the complex in positionB and returns the imaginary part of that operation.
	 * @param complexA complex number to operate
	 * @param positionB index of the complex number in the array
	 * @return double with the imaginary part of the multiplication
	 */
	double multiplyComplexWithPositionAndReturnImag(Complex complexA, int positionB);
	/**
	 * Swaps the position of two values.
	 * @param positionA first position to swap
	 * @param positionB second position to swap
	 */
	void swapPositions(int positionA, int positionB);
	/**
	 * adds the complexB to the complex in positionA and stores in positionA.
	 * @param positionA position of the receiving complex
	 * @param complexB complex value to add
	 */
	void add(int positionA, Complex complexB);
	/**
	 * adds the complex value of real and imaginary to the complex in positionA and stores in positionA.
	 * @param positionA position of the receiving complex
	 * @param complexB complex value to add
	 */
	void add(int positionA, double real, double imaginary);
	/**
	 * substracts the complexB to the complex in positionA and stores in positionA.
	 * @param positionA position of the receiving complex
	 * @param complexB complex value to add
	 */
	void substract(int positionA, Complex complexB);
	/**
	 * substracts the complex value of real and imaginary to the complex in positionA and stores in positionA.
	 * @param real real part of the complex number
	 * @param imaginary imaginary part of the complex number
	 * @param complexB complex value to add
	 */
	void substract(int positionA, double real, double imaginary);
	/**
	 * multiplies the complex in positionA with complexB and stores in positionA.
	 * @param positionA position of the receiving complex
	 * @param complexB complex value to add
	 */
	void multiply(int positionA, Complex complexB);
	/**
	 * multiplies the complex in positionA with  the real value realB and stores in positionA.
	 * @param positionA position of the receiving complex
	 * @param complexB complex value to add
	 */
	void multiply(int positionA, double realB);

	double absolute(int positionA);
}
