package org.lossywav.dsp;
//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007, 2008 Nick Currie, Copyleft.
//    Copyright (C) 2008, C# port by Greg Chudov.
//    Copyright (C) 2008, Java port by Josep Maria Antoln
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> yahoo <dot> com
/**
 * Implementation of Complex buffer with a single array of doubles (to speedup access of real and imag. parts).
 * @author JosepMa
 *
 */
public class DoubleArrayFastBuffer implements ComplexBuffer {
	double[] array;
	public DoubleArrayFastBuffer(int bufferSize) {
		array = new double[bufferSize];
	}
	
	public Complex getComplex(int position) {
		final int realPos = position * 2;
		return new Complex(array[realPos],array[realPos+1]);
	}

	
	public double getImag(int position) {
		return array[position * 2 +1];
	}

	
	public double getReal(int position) {
		return array[position * 2];
	}

	
	public void setComplex(int position, Complex complex) {
		final int realPos = position * 2;
		array[realPos] = complex.re();
		array[realPos+1] = complex.im();
	}
	
	public void setComplex(int position, double real, double img) {
		final int realPos = position * 2;
		array[realPos] = real;
		array[realPos+1] = img;
	}
	
	public void setImag(int position, double imag) {
		array[position * 2 + 1] = imag;
	}

	
	public void setReal(int position, double real) {
		array[position * 2] = real;
	}
	
	public void add(int positionA, double real, double imaginary) {
		final int realPos = positionA *2;
		array[realPos] += real;
		array[realPos+1] += imaginary;
	}
	
	public void add(int positionA, Complex complexB) {
		final int realPos = positionA * 2;
		array[realPos] += complexB.re();
		array[realPos+1] += complexB.im();
	}

	
	public void substract(int positionA, Complex complexB) {
		final int realPos = positionA * 2;
		array[realPos] -= complexB.re();
		array[realPos+1] -= complexB.im();
	}
	
	public void substract(int positionA, double real, double imaginary) {
		final int realPos = positionA *2;
		array[realPos] -= real;
		array[realPos+1] -= imaginary;
	}
	
	public void multiply(int positionA, Complex complexB) {
		final int realPos = positionA * 2;
		final double reTmp = array[realPos];
		final double imgTmp = array[realPos +1];

		array[realPos] = reTmp * complexB.re() - imgTmp * complexB.im();
		array[realPos + 1] = reTmp * complexB.im() + imgTmp * complexB.re();
	}
	
	public void multiply(int positionA, double realB) {
		final int realPos = positionA * 2;
		array[realPos] *= realB;
		array[realPos+1] *= realB;
	}

	
	public Complex sumPositions(int positionA, int positionB) {
		final int realPosA = positionA *2;
		final int realPosB = positionB * 2;
		final double re = array[realPosA] + array[realPosB];
		final double img = array[realPosA + 1] + array[realPosB + 1];

		return new Complex(re, img);
	}
	
	public void sumPositionsInA(int positionA, int positionB) {
		final int realPosA = positionA *2;
		final int realPosB = positionB * 2;
		array[realPosA] += array[realPosB];
		array[realPosA + 1] += array[realPosB + 1];
	}

	
	public Complex substractPositions(int positionA, int positionB) {
		final int realPosA = positionA *2;
		final int realPosB = positionB * 2;
		final double re = array[realPosA] - array[realPosB];
		final double img = array[realPosA + 1] - array[realPosB + 1];

		return new Complex(re, img);
	}
	
	public void substractPositionsInA(int positionA, int positionB) {
		final int realPosA = positionA *2;
		final int realPosB = positionB * 2;
		array[realPosA] -= array[realPosB];
		array[realPosA + 1] -= array[realPosB + 1];
	}

	
	public Complex multiplyPositions(int positionA, int positionB) {
		final int realPosA = positionA *2;
		final int realPosB = positionB * 2;
		final double reA = array[realPosA];
		final double imgA = array[realPosA +1];
		final double reB = array[realPosB];
		final double imgB = array[realPosB +1];

		return new Complex(reA * reB - imgA * imgB,
							reA * imgB + imgA * reB);
	}
	
	public void multiplyPositionsInA(int positionA, int positionB) {
		final int realPosA = positionA *2;
		final int realPosB = positionB * 2;
		final double reA = array[realPosA];
		final double imgA = array[realPosA +1];
		final double reB = array[realPosB];
		final double imgB = array[realPosB +1];

		array[realPosA] = reA * reB - imgA * imgB;
		array[realPosA + 1] = reA * imgB + imgA * reB;
	}
	
	public double multiplyComplexWithPositionAndReturnReal(Complex complexA,
			int positionB) {
		final int realPos = positionB *2;
       return  complexA.re() * array[realPos] - complexA.im() * array[realPos + 1];
	}
	
	public double multiplyComplexWithPositionAndReturnImag(Complex complexA,
			int positionB) {
		final int realPos = positionB *2;
        return complexA.re() * array[realPos + 1] + complexA.im() * array[realPos];

	}
	
	public void swapPositions(int positionA, int positionB) {
		final int realPosA = positionA *2;
		final int realPosB = positionB * 2;
		final double reTmp = array[realPosA];
		final double imgTmp = array[realPosA +1];
		array[realPosA] = array[realPosB];
		array[realPosA + 1] = array[realPosB + 1];
		array[realPosB] = reTmp;
		array[realPosB + 1] = imgTmp;

	}
	
	public double absolute(int positionA) {
		final int realPosA = positionA *2;
		final double re = array[realPosA];
		final double im = array[realPosA + 1];
		return Math.sqrt(re*re + im*im);
	}
}
