package org.lossywav.dsp;
//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007, 2008 Nick Currie, Copyleft.
//    Copyright (C) 2008, C# port by Greg Chudov.
//    Copyright (C) 2008, Java port by Josep Maria Antoln
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> yahoo <dot> com
import org.lossywav.util.Constants;
/**
 * Inplace FFT signal analyzer for power of two sample sizes.
 * @author JosepMa
 *
 */
public class InplaceFFT {
	/**
	 * angle buffer in a ComplexBuffer interface.
	 */
	private static final ComplexBuffer[] angleBuffers;
	/**
	 * Buffer of reversed number bits. ex: 4 decimal = 0100 binary -> 0010 reversed (The stored reversion uses 32 bits )
	 */
	private static final int[] reversedBits;
	/**
	 * audio signal in a ComplexBuffer interface.
	 */
	private final ComplexBuffer signal;
	/**
	 * Static constructor.
	 */
	static {
		angleBuffers = new ComplexArrayBuffer[Constants.MAX_FFT_BITWIDTH+1];
		calculateLookupTable();

		reversedBits = new int[1 << Constants.MAX_FFT_BITWIDTH];
		calculateReversedBitsTable();
	}
	/**
	 * Constructor.
	 */
	public InplaceFFT() {
		signal = new DoubleArrayFastBuffer(1 << Constants.MAX_FFT_BITWIDTH);
	}

    /**
     * Prefills the angle lookup buffer tables used in the fftComplex calculation.
     */
    private static void calculateLookupTable() {
    	for (int bits = 1; bits <= Constants.MAX_FFT_BITWIDTH; bits++) {
			final int blockend = 1 << bits;
			angleBuffers[bits] = new ComplexArrayBuffer(blockend);
			final double factor = Math.PI / blockend;
	    	// The value for the position 0 of complex is always 1, so it can be ommited.
			for (int j = 1; j < blockend; j++) {
				angleBuffers[bits].setComplex(j, Math.cos(-j * factor),	
												Math.sin(-j * factor));
			}
    	}
    }
    private static void calculateReversedBitsTable() {
		final int halfLength = Constants.MAX_FFT_LENGTH >> 1;

		reversedBits[0] = 0;
		reversedBits[1] = halfLength;
		for (int i = 1; i < halfLength; i++) {
			final int rb2n = reversedBits[i] >> 1;
			reversedBits[i << 1] = rb2n;
			reversedBits[(i << 1) + 1] = rb2n | halfLength;
		}
    }
	/**
	 * Obtaion access to the signal.
	 * @return ComplexBuffer element
	 */
	public ComplexBuffer getX() {
		return signal;
	}
	/**
	 * Computes the Real number FFT of the signal.
	 * @param numberOfBits  power of two exponent equivalent to the number of samples
	 */
	public void fftReal(int numberOfBits) {

		final int n = 1 << numberOfBits;
		final int nHalf = 1 << (numberOfBits - 1);

		fftComplex(numberOfBits);

		signal.setComplex(n, signal.getReal(0) - signal.getImag(0), 0);
		signal.setComplex(0, signal.getReal(0) + signal.getImag(0), 0);
		final ComplexBuffer currentAngleBuffer = angleBuffers[numberOfBits];
		for (int j = 1; j < nHalf; j++) {
			final int k = n - j;
			final double realA = currentAngleBuffer.getReal(j);
			final double imagA = currentAngleBuffer.getImag(j);

			final double realJ = signal.getReal(j);
			final double imagJ = signal.getImag(j);
			final double realK = signal.getReal(k);
			final double imagK = signal.getImag(k);

            final double realX0 = realJ + realK;
            final double imagX0 = imagJ + imagK;
            final double realX1 = realJ - realK;
            final double imagX1 = imagJ - imagK;
            final double realV1 = realA * imagX0 + imagA * realX1;
            final double imagV1 = realA * realX1 - imagA * imagX0;

            final double newRealJ = (realX0 + realV1) *0.5;
            final double newImagJ = (imagX1 - imagV1) *0.5;
            final double newRealK = (realX0 - realV1) *0.5;
            final double newImagK = -(imagX1 + imagV1) *0.5;
            signal.setComplex(j, newRealJ, newImagJ);
            signal.setComplex(k, newRealK, newImagK);
		}
	}
	/**
	 * compute the complex FFT of the signal, assuming its length is a power of 2.
	 * @param powerOfTwoBits  power of two factor equivalent to the number of samples
	 */
    public void fftComplex(final int powerOfTwoBits) {
        final int numberOfSamples = 1 << powerOfTwoBits;

        swapByPositionBitsReverse(numberOfSamples, powerOfTwoBits);

		for (int blockbitlen = 0; blockbitlen < powerOfTwoBits; blockbitlen++)
		{
			final int halfSamples = 1 << blockbitlen;
			final int blocksize = halfSamples  << 1;
			final ComplexBuffer currentAngle = angleBuffers[blockbitlen];
			for( int i = 0; i < numberOfSamples; i+=blocksize) {
				// On the first pass, angle*signal[k] equals signal[k], so I simplify the calculation here.
				double realT, imagT;
				int k = i + halfSamples;
				realT = signal.getReal(k);
				imagT = signal.getImag(k);
                signal.setComplex(k, signal.getReal(i) - realT, signal.getImag(i) - imagT);
                signal.add(i, realT, imagT);

				for (int n = 1, j = i+1; n < halfSamples; n++, j++) {
					k = j + halfSamples;
	            	//this getComplex uses a preallocated value. no creation takes place.
            		final Complex angle = currentAngle.getComplex(n);
            		realT = signal.multiplyComplexWithPositionAndReturnReal(angle, k);
            		imagT = signal.multiplyComplexWithPositionAndReturnImag(angle, k);
            		// This is the equivalent line of the above two ones
            		// final Complex tao = angle.times(signal.getComplex(k));
                    signal.setComplex(k, signal.getReal(j) - realT, signal.getImag(j) - imagT);
                    signal.add(j, realT, imagT);
				}
			}
        }
    }

    /**
     * Swaps the positions of the values of the array to prepare it for the FFT analysis.
     * The position is determined by reversing the bits of the position index ( 00001 -> 10000 )
     * If the result of the FFT was to be written to a secondary array, this could be implemented
     * as a lookup table instead of actually moving the data.
     * @param numberOfSamples number of samples to reverse
     * @param numberOfBits  power of two factor equivalent to this numberOfSamples
     */
    private void swapByPositionBitsReverse(int numberOfSamples, int numberOfBits) {
    	final int shift = Constants.MAX_FFT_BITWIDTH - numberOfBits;
        for (int k = 0; k < numberOfSamples; k++) {
        	final int j = reversedBits[k] >> shift;
            if (j > k) {
            	signal.swapPositions(j, k);
            }
        }
    }
}
