package org.lossywav.frontend;
//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//              Noise shaping coefficients by Sebastian Gesemann;
//
//    Copyright (C) 2007, 2008 Nick Currie, Copyleft.
//    Copyright (C) 2008, C# port by Greg Chudov.
//    Copyright (C) 2008, Java port by Josep Maria Antoln
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> yahoo <dot> com
import java.io.File;
import java.io.IOException;
import java.text.ParseException;
import java.util.Arrays;
import java.util.Iterator;

import org.lossywav.algorithm.LossyWav;
import org.lossywav.audiocodecs.WavReader;
import org.lossywav.audiocodecs.WavWriter;
import org.lossywav.util.Constants;
import org.lossywav.util.LossyWavException;

public class JLossyWavConsole {
	private LossyWav lossyWav;
	private StringBuilder parameterString = new StringBuilder();
	private String inputFile = null;
	private String stdinName = null;
	private String outputDir = null;
	private boolean merge = false;
	private boolean forceOverwrite = false;
	private boolean writeCorrection = false;

	public static void main(String... args)
	{
		if (args.length == 0 ) {
			ConsoleUtil.PrintStringArray(ConstConsole.LOSSY_WAV_HEAD);
			ConsoleUtil.PrintStringArray(ConstConsole.LOSSYWAV_GPL);
			ConsoleUtil.PrintStringArray(ConstConsole.LOSSYWAV_SYNTAX);
			System.exit(-1);
		}
		if (ConsoleUtil.isHelpParam(args[0])) {
			System.exit(1);
		}
		
		JLossyWavConsole jlossyWavConsole = new JLossyWavConsole();
		int retVal = jlossyWavConsole.run(args);
		System.exit(retVal);
	}

	public int run(String... args) {
		lossyWav = new LossyWav();
		ConsoleUtil.PrintStringArray(ConstConsole.LOSSY_WAV_HEAD, System.out);
		try {
			parseOptions(Arrays.asList(args).iterator());
		} catch (ParseException e) {
			System.err.println(e.getMessage());
			return -2;
		} catch (LossyWavException e) {
			System.err.println(e.getMessage());
			return -2;
		}
		//TODO: Missing the verification of incompatible switches and other details.

		WavReader audioSource = null;
		WavWriter audioDest = null;
		WavWriter lwcdfDest = null;
		try	{
			
			if ( inputFile != null ) { 
				audioSource = new WavReader(inputFile);
			} else {
				//TODO: implement
				if (stdinName != null){
					inputFile = stdinName;
				} else {
					inputFile = "stdin";
				}
				throw new ParseException("stdin input not supported yet",0);
			}
			String outputFile;
			if (outputDir != null) {
				outputFile = outputDir.concat(inputFile);
			} else {
				outputFile = inputFile;
			}
			if ( merge ) {
				//TODO: implement
				lossyWav.merge(System.out);
				throw new ParseException("Merge not supported yet",0);
			}
			else {
				if (!forceOverwrite) {
					File file = new File(outputFile.concat(".lossy.wav"));
					if (file.exists()) {
						System.err.println("Error: The file ".concat(file.getName()).concat(" already exists."));
						return -31;
					}
					if (writeCorrection) {
						File corr = new File(outputFile.concat(".lwcdf.wav"));
						if (corr.exists()) {
							System.err.println("Error: The file ".concat(file.getName()).concat(" already exists."));
							return -31;
						}
					}
				}
				audioDest = new WavWriter(outputFile.concat(".lossy.wav"), (short) audioSource.getBitsPerSample(),
								(short) audioSource.getChannelCount(), audioSource.getSampleRate());
				if (writeCorrection) {
					lwcdfDest = new WavWriter(outputFile.concat(".lwcdf.wav"), (short) audioSource.getBitsPerSample(),
								(short) audioSource.getChannelCount(), audioSource.getSampleRate());
				}

				lossyWav.setParameterString(parameterString.toString());
				lossyWav.setAudioSource(audioSource);
				lossyWav.setAudioDest(audioDest);
				lossyWav.setCorrectionDest(lwcdfDest);
				lossyWav.process(System.out);
			}
		}
		catch (final Exception ex) {
			System.err.println("Error: " + ex.getMessage());
			ex.printStackTrace();
		}
		finally {
			try {
				if (audioSource != null ) {
					audioSource.close();
				}
				if (audioDest != null) {
					audioDest.close();
				}
				if (lwcdfDest != null) {
					lwcdfDest.close();
				}
			} catch (final IOException e) {
				e.printStackTrace();
			}
		}
		return 0;
	}
	
	private void parseOptions(Iterator<String> args) throws ParseException, LossyWavException {
		while ( args.hasNext() ) {
			String arg = args.next();
			parameterString.append(" ");
			if (parseQualityParam(arg, args)) {
				continue;
			}
			if (parseSystemParam(arg, args)) {
				continue;
			}
			if (parseOperationParam(arg, args)) {
				continue;
			}
			if (parseAdvancedParam(arg, args)) {
				continue;
			}
			// Name always comes last
			if (! args.hasNext()) {
				if ("-".equals(arg)) {
					inputFile = null;
					continue;
				}
				else if ('-' != arg.charAt(0)) {
					//TODO: verify file exists.
					inputFile = arg;
					continue;
				}
			}
			throw new ParseException(ConstConsole.ERRORS_PARSING_UNKNOWN + arg, 0);
		}
	}
	private boolean parseQualityParam(String current, Iterator<String> params) throws ParseException {
		if ("-q".equals(current) || "--quality".equals(current)) {
			if ( !params.hasNext()) {
				throw new ParseException(ConstConsole.ERRORS_VAL_MISSING_FOR + current, 0);
			}
			double val;
			try {
				val = Double.parseDouble(params.next());
			} catch(NumberFormatException e) {
				throw new ParseException(ConstConsole.ERRORS_NONNUMBER + current, 0);
			}
			ConsoleUtil.checkValidRange(current, val, Constants.MIN_QUALITY, Constants.MAX_QUALITY);
			lossyWav.setQuality(val);
			parameterString.append("--quality ").append(val);
			return true;
		}
		for (int i = 0; i < ConstConsole.NUM_QUALITY_SYNONYMS; i++) {
			final String synLong = ConstConsole.QUALITY_SYNONYMS_LONG[i];
			final String synShort = ConstConsole.QUALITY_SYNONYMS_SHORT[i];
			if ( synLong.equals(current) || synShort.equals(current)) {
				lossyWav.setQuality(ConstConsole.QUALITY_SYNONYMS_VALS[i]);
				parameterString.append(synLong);
				return true;
			}
		}
		if ("--maxclips".equals(current)) {
			if ( !params.hasNext()) {
				throw new ParseException(ConstConsole.ERRORS_VAL_MISSING_FOR + current, 0);
			}
			int val;
			try {
				val = Integer.parseInt(params.next());
			} catch(NumberFormatException e) {
				throw new ParseException(ConstConsole.ERRORS_NONNUMBER + current, 0);
			}
			ConsoleUtil.checkValidRange(current, val, Constants.MIN_CLIPS, Constants.MAX_CLIPS);
			lossyWav.setMaxClipsAllowed(val);
			parameterString.append(current).append(" ").append(val);
			return true;
		}
		return false;
	}
	private boolean parseSystemParam(String current, Iterator<String> params) throws ParseException {
		if ("--stdout".equals(current)) {
			//TODO:implement
			return true;
		}
		if ("-B".equals(current) || "--below".equals(current)) {
			Thread.currentThread().setPriority(Thread.MIN_PRIORITY);
			return true;
		}
		if ("--low".equals(current)) {
			Thread.currentThread().setPriority(Thread.MIN_PRIORITY);
			return true;
		}
		return false;
	}

	private boolean parseAdvancedParam(String current, Iterator<String> params) throws ParseException, LossyWavException {
		if ("-s".equals(current) || "--shaping".equals(current)) {
			if ( !params.hasNext()) {
				throw new ParseException(ConstConsole.ERRORS_VAL_MISSING_FOR + current, 0);
			}
			double val;
			try {
				val = Double.parseDouble(params.next());
			} catch(NumberFormatException e) {
				throw new ParseException(ConstConsole.ERRORS_NONNUMBER + current, 0);
			}
			ConsoleUtil.checkValidRange(current, val, Constants.MIN_SHAPING, Constants.MAX_SHAPING);
			lossyWav.setShapingFactor(val);
			parameterString.append("--shaping ").append(val);
			return true;
		}
		if ("--scale".equals(current)) {
			if ( !params.hasNext()) {
				throw new ParseException(ConstConsole.ERRORS_VAL_MISSING_FOR + current, 0);
			}
			double val;
			try {
				val = Double.parseDouble(params.next());
			} catch(NumberFormatException e) {
				throw new ParseException(ConstConsole.ERRORS_NONNUMBER + current, 0);
			}
			ConsoleUtil.checkValidRange(current, val, Constants.MIN_SCALING, Constants.MAX_SCALING);
			lossyWav.setScalingFactor(val);
			parameterString.append("--scale ").append(val);
			return true;
		}
		if ("-l".equals(current) || "--limit".equals(current)) {
			if ( !params.hasNext()) {
				throw new ParseException(ConstConsole.ERRORS_VAL_MISSING_FOR + current, 0);
			}
			int val;
			try {
				val = Integer.parseInt(params.next());
			} catch(NumberFormatException e) {
				throw new ParseException(ConstConsole.ERRORS_NONNUMBER + current, 0);
			}
			ConsoleUtil.checkValidRange(current, val, Constants.MIN_UPPER_FREQUENCY_LIMIT, Constants.MAX_UPPER_FREQUENCY_LIMIT);
			lossyWav.setFrequencyLimit(val);
			parameterString.append("--limit ").append(val);
			return true;
		}
		if ("-a".equals(current) || "--analyses".equals(current)) {
			if ( !params.hasNext()) {
				throw new ParseException(ConstConsole.ERRORS_VAL_MISSING_FOR + current, 0);
			}
			int val;
			try {
				val = Integer.parseInt(params.next());
			} catch(NumberFormatException e) {
				throw new ParseException(ConstConsole.ERRORS_NONNUMBER + current, 0);
			}
			ConsoleUtil.checkValidRange(current, val, Constants.MIN_PARAM_ANALYSIS, Constants.MAX_PARAM_ANALYSIS);
			lossyWav.setAnalysisNumber(val);
			parameterString.append("--analyses ").append(val);
			return true;
		}
		if ("-i".equals(current) || "--impulse".equals(current)) {
			lossyWav.setForceFFT32(true);
			return true;
		}
		if ("-m".equals(current) || "--midside".equals(current)) {
			lossyWav.setUseMidSide(true);
			parameterString.append("--midside ");
			return true;
		}		
		if ("--linkchannels".equals(current)) {
			lossyWav.setLinkChannelBitsToRemove(true);
			parameterString.append("--linkchannels ");
			return true;
		}		
		if ("-X".equals(current) || "--sortspread".equals(current)) {
			lossyWav.setSortSpread(true);
			parameterString.append("--sortspread ");
			return true;
		}		
		if ("-U".equals(current) || "--underlap".equals(current)) {
			if ( !params.hasNext()) {
				throw new ParseException(ConstConsole.ERRORS_VAL_MISSING_FOR + current, 0);
			}
			int val;
			try {
				val = Integer.parseInt(params.next());
			} catch(NumberFormatException e) {
				throw new ParseException(ConstConsole.ERRORS_NONNUMBER + current, 0);
			}
			ConsoleUtil.checkValidRange(current, val, Constants.MIN_UNDERLAP, Constants.MAX_UNDERLAP);
			lossyWav.setUnderlap(val);
			parameterString.append("--underlap ").append(val);
			return true;
		}
		if ("--deltants".equals(current)) {
			if ( !params.hasNext()) {
				throw new ParseException(ConstConsole.ERRORS_VAL_MISSING_FOR + current, 0);
			}
			double val;
			try {
				val = Double.parseDouble(params.next());
			} catch(NumberFormatException e) {
				throw new ParseException(ConstConsole.ERRORS_NONNUMBER + current, 0);
			}
			ConsoleUtil.checkValidRange(current, val, Constants.MIN_DELTANTS, Constants.MAX_DELTANTS);
			lossyWav.setDeltaNTS(val);
			parameterString.append("--deltants ").append(val);
			return true;
		}
		if ("--deltasnr".equals(current)) {
			if ( !params.hasNext()) {
				throw new ParseException(ConstConsole.ERRORS_VAL_MISSING_FOR + current, 0);
			}
			double val;
			try {
				val = Double.parseDouble(params.next());
			} catch(NumberFormatException e) {
				throw new ParseException(ConstConsole.ERRORS_NONNUMBER + current, 0);
			}
			ConsoleUtil.checkValidRange(current, val, Constants.MIN_DELTASNR, Constants.MAX_DELTASNR);
			lossyWav.setDeltaSNR(val);
			parameterString.append("--deltasnr ").append(val);
			return true;
		}		
		return false;
	}
	private boolean parseOperationParam(String current, Iterator<String> params) throws ParseException {
		if ("-c".equals(current) || "--check".equals(current)) {
			lossyWav.setCheckIfProcessed(true);
			return true;
		}
		if ("-M".equals(current) || "--merge".equals(current)) {
			merge = true;
			return true;
		}
		if ("-f".equals(current) || "--force".equals(current)) {
			forceOverwrite = true;
			return true;
		}
		if ("-C".equals(current) || "--correction".equals(current)) {
			writeCorrection = true;
			return true;
		}

	    if ("--stdinname".equals(current)) {
			if ( !params.hasNext()) {
				throw new ParseException(ConstConsole.ERRORS_VAL_MISSING_FOR + current, 0);
			}
			stdinName = params.next();
			return true;
		}
		if ("-o".equals(current) ||"--outdir".equals(current)) {
			if ( !params.hasNext()) {
				throw new ParseException(ConstConsole.ERRORS_VAL_MISSING_FOR + current, 0);
			}
			outputDir = params.next();
			if (outputDir.endsWith(":") || !outputDir.endsWith("\\")) {
				outputDir = outputDir.concat("\\");
			}
			return true;
		}
		return false;
	}
		/*

  function check_output : boolean;
  begin
    check_output:=false;

    if (parmchar='d') or (current_parameter='--detail') then
    begin
      parmerror:='detail';
      if (detail_int) then
        parmerror_multiple_selection;
      detail_int:=true;
      check_output:=true;
    end;

    if (current_parameter='--blockdist') then
    begin
      parmerror:='block lsb distribution';
      if (blockdist_int) then
        parmerror_multiple_selection;
      blockdist_int:=true;
      check_output:=true;
    end;

    if (current_parameter='--sampledist') then
    begin
      parmerror:='sample lsb distribution';
      if (sampledist_int) then
        parmerror_multiple_selection;
      sampledist_int:=true;
      check_output:=true;
    end;

    if (parmchar='Q') or (current_parameter='--quiet') then
    begin
      parmerror:='reduced output';
      if not verbosity_int then
        parmerror_multiple_selection;
      verbosity_int:=false;
      check_output:=true;
    end;

    if (current_parameter='--silent') then
    begin
      parmerror:='silent running';
      if silent_int=true then
        parmerror_multiple_selection;
      silent_int:=true;
      check_output:=true;
    end;

    if (parmchar='n') or (current_parameter='--nowarnings') then
    begin
      parmerror:='no warnings';
      if not warnings_int then
        parmerror_multiple_selection;
      warnings_int:=false;
      check_output:=true;
    end;

  function check_file_output : boolean;
  begin
    check_file_output:=false;

    if (parmchar='w') or (current_parameter='--writetolog') then
    begin
      parmerror:='write log output';
      if writetolog_int then
        parmerror_multiple_selection;
      writetolog_int:=true;
      check_file_output:=true;
    end;
 
		 */
}
